/*
 * GWT - General Windowing Toolkit
 *
 * Copyright (C) 1999 Rodolphe Ortalo - ortalo@laas.fr
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: visible.c,v $
 * Revision 1.3  2000/03/19 19:32:46  ortalo
 * Some debugging and modifications (wrt OpenAmulet mainly).
 *
 * Revision 1.2  1999/05/20 20:34:23  ortalo
 * Several bugs corrections and small improvements.
 *
 * Revision 1.1  1999/05/13 20:01:38  ortalo
 * New version of libgwt. Now using LibGII for input management.
 * Some drawing functions implemented. Visibility management should
 * be rather okay. Still some work to do, but now, your eyes can
 * look at some drawings... not only code. ;-)
 *
 */

#include "internal.h"

/************************************************
 *  Convenience visibility management functions *
 ************************************************/
/*
 * Computes the union of the geometry of all the children
 * of the given window. Do not include any hidden children.
 *
 * MODIFIES: The given region 'res' is modified !
 */
static void gwt_compute_all_children_region(const gwt_window_t win,
					    const gwt_region_t res)
{
  gwt_window_t child;
  /* Zeroes the region */
  gwt_region_init(res, 0, 0, 0, 0);
  /* Iterate over children */
  child = GWT_Z_BOTTOM_CHILD(win);
  while (child != NULL)
    {
      if (!gwtIsHidden(child))
	/* _explicitly_ hidden: win should be visible if we are called */
	gwt_region_union(res, child->geometry);
      child = GWT_Z_UP(child);
    }
  return;
}
/*
 * Computes union of all the siblings of the given window
 * that are before (on top of) it in the Z-order.
 *
 * MODIFIES: The given region 'res' is modified !
 */
static void gwt_compute_covering_siblings_region(const gwt_window_t win,
						 const gwt_region_t res)
{
  gwt_window_t sibling;
  /* Zeroes the region */
  gwt_region_init(res, 0, 0, 0, 0);
  /* Iterate 'upwards' over siblings */
  sibling = GWT_Z_UP(win);
  while (sibling != NULL)
    {
      if (!gwtIsHidden(sibling))
	gwt_region_union(res, sibling->geometry);
      sibling = GWT_Z_UP(sibling);
    }
  return; /* Do not fail */
}

/******************************************************
 *  Implementation of visibility management functions *
 ******************************************************/
/*
 * Computes the potentially visible region of this window.
 *
 * REQUIRES: The potentially visible region of the parent
 *   must be up-to-date (or there should be no parent)!
 *   The geometry of the siblings must be up-to-date also.
 * MODIFIES: The potentially visible region of 'win' is modified !
 */
static void gwt_potentially_visible_update(const gwt_window_t win)
{
  gwt_region_t up_siblings; /* temporary region */
  /* Initializes with the geometry */
  gwt_region_copy(win->potentially_visible, win->geometry);
  if (win->wparent != NULL)
    {
      /* Intersects with the parent potentially visible region */
      gwt_region_intersection(win->potentially_visible,
			      win->wparent->potentially_visible);
      /* Get upwards siblings region */
      gwt_region_create(&up_siblings);
      gwt_compute_covering_siblings_region(win, up_siblings);
      /* And removes that area */
      gwt_region_difference(win->potentially_visible, up_siblings);
      gwt_region_destroy(up_siblings); /* Releases memory */
    }
  return;
}
/*
 * Computes the visible region of this window
 *
 * REQUIRES: The potentially visible region of the window
 *   must be up-to-date!
 * MODIFIES: The visible region of 'win' is modified !
 */
static void gwt_visible_update(const gwt_window_t win)
{
  gwt_region_t children; /* temporary region */
  /* Initializes with the potentially visible region */
  gwt_region_copy(win->visible, win->potentially_visible);
  /* Get children region */
  gwt_region_create(&children);
  gwt_compute_all_children_region(win, children);
  /* And removes that area */
  gwt_region_difference(win->visible, children);
  gwt_region_destroy(children); /* Releases memory */
  return;
}
/*
 * This function updates the potentially_visible and
 * visible regions of a window and its children.
 * (via Recursion).
 *
 * MODIFIES:
 *   - the potentially_visible, visible and drawing regions of the
 *   window
 *   - the potentially_visible, visible and drawing regions of
 *   every child (and sub-child) - via recursion.
 */
static void gwt_visibility_update(const gwt_window_t changed)
{
  gwt_window_t child;
  /* TODO: Breaks the recursion if window is hidden (nothing done) */
  /* TODO: Check that this behavior is correct wrt/ hidden state */
  if (gwtIsHidden(changed)) return;
  /* Update the changed window */
  gwt_potentially_visible_update(changed);
  gwt_visible_update(changed);
  gwt_drawing_update(changed);
  /* Iterate over all children */
  child = GWT_Z_BOTTOM_CHILD(changed);
  while (child != NULL)
    {
      gwt_visibility_update(child);
      child = GWT_Z_UP(child);
    }
  return;
}
/*********************************************
 *      Update Internal API entry point      *
 *********************************************/
/*
 * Update a window _family_ visible region(s) after a
 * geometry or hide modification of one of its member
 *
 * REQUIRES: the parent potentially visible region should
 *   be up-to-date. The geometry should be up-to-date.
 * MODIFIES: - the visible region of the parent;
 *   - the visibility regions of the window (and children);
 *   - the visibility regions of every sibling BELOW it
 *     (and children)
 */
void gwt_propagate_geometry_update(const gwt_window_t changed)
{
  gwt_window_t sibling;
  /*  First, we update the visible region of the
     parent window (it may have changed since one child
     changed). Parent may not exist sometimes. */
  if (changed->wparent != NULL)
    {
      gwt_visible_update(changed->wparent);
      gwt_drawing_update(changed->wparent);
    }
  /* Then we should update the downwards siblings */
  sibling = GWT_Z_DOWN(changed);
  while (sibling != NULL)
    {
      gwt_visibility_update(sibling);
      sibling = GWT_Z_DOWN(sibling);
    }
  /* Finally, we need to update the window itself */
  gwt_visibility_update(changed);
  return;
}
/*
 * Update a window _family_ visible region(s) after a
 * hide/show, creation/destruction or Z-order modification
 * of one child.
 *
 * REQUIRES: the window potentially visible region should
 *   be up-to-date. Its geometry should be up-to-date.
 * MODIFIES: - the visible region of the window;
 *   - the visibility regions of ALL the children.
 */
void gwt_propagate_child_update(const gwt_window_t window)
{
  gwt_window_t child;
  /*  First, we update the visible region of the window
      (it may have changed since one child changed). */
  gwt_visible_update(window);
  gwt_drawing_update(window);
  /* Iterate over all children */
  child = GWT_Z_BOTTOM_CHILD(window);
  while (child != NULL)
    {
      gwt_visibility_update(child);
      child = GWT_Z_UP(child);
    }
  return;
}
/*
 * Update the implicit hidden flag of a window
 *
 * REQUIRES: - the window must have its hidden flag
 *  up-to-date.
 *   - the parent window must have its implicitly hidden
 *  flag up-to-data
 * MODIFIES: the implicitly hidden flag of the window
 *  and ALL its children.
 */
void gwt_implicitly_hidden_update(const gwt_window_t win)
{
  gwt_window_t curr;

  /* Update this window flag */
  win->implicitly_hidden = win->wparent->implicitly_hidden
    || win->wparent->hidden;
  /* Recursively update all windows */
  curr = GWT_Z_BOTTOM_CHILD(win);
  while (curr != NULL)
    {
      gwt_implicitly_hidden_update(curr);
      curr = GWT_Z_UP(curr);
    }
  return;
}
